/****************************************************************************
 ****************************************************************************/

#include "jwm.h"

static const int STACK_BLOCK_SIZE = 2;

ClientNode *nodes = NULL;
ClientNode *nodeTail = NULL;

static ClientNode *activeClient;

static Window *stack;
static int stackSize;
static int clientCount;

static void Reparent(ClientNode *np);
 
#ifdef USE_MWM_HINTS
static PropMwmHints *GetMwmHints(Window w);
#endif
static void SendClientMessage(ClientNode *np, Atom type, Atom message);
static void GetBorderOffsets(ClientNode *np, int *north, int *west);
static void PlaceWindow(ClientNode *np, int alreadyMapped);
static void Gravitate(ClientNode *np, int negate);
static void UpdateState(ClientNode *np);
static void MinimizeTransients(ClientNode *np);
static void CheckShape(ClientNode *np);

/****************************************************************************
 ****************************************************************************/
void LoadClients() {
	XWindowAttributes attr;
	Window rootReturn, parentReturn, *childrenReturn;
	unsigned int childrenCount;
	unsigned int x;

	XFlush(display);
	XGrabServer(display);

	stackSize = STACK_BLOCK_SIZE;
	stack = Allocate(sizeof(Window) * stackSize);
	clientCount = 0;
	nodes = NULL;
	nodeTail = NULL;
	activeClient = NULL;
	currentDesktop = 0;

	XQueryTree(display, rootWindow, &rootReturn, &parentReturn,
		&childrenReturn, &childrenCount);

	for(x = 0; x < childrenCount; x++) {
		if(XGetWindowAttributes(display, childrenReturn[x], &attr)) {
			if(attr.override_redirect == False
				&& attr.map_state == IsViewable) {
				AddClientWindow(childrenReturn[x], 1);
			}
		}
	}

	XFree(childrenReturn);

	XSync(display, True);
	RestackClients();

	XUngrabServer(display);
	XFlush(display);

}

/****************************************************************************
 ****************************************************************************/
void AddClientWindow(Window w, int alreadyMapped) {
	XWindowAttributes attr;
	XWMHints *wmhints;
	ClientNode *np;
	int startMapped;
	long lvalue;

	Assert(w != None);

	if(w == trayWindow || w == rootWindow) {
		return;
	}

	XGetWindowAttributes(display, w, &attr);

	if(attr.override_redirect == True) {
		return;
	}

	np = Allocate(sizeof(ClientNode));
	memset(np, 0, sizeof(ClientNode));
	np->prev = NULL;
	if(nodes) {
		nodes->prev = np;
	} else {
		nodeTail = np;
	}
	np->next = nodes;
	nodes = np;
	np->window = w;
	np->groupLeader = None;
	np->owner = None;
	np->desktop = currentDesktop;

	XFetchName(display, w, &np->name);

	np->x = attr.x;
	np->y = attr.y;
	np->width = attr.width;
	np->height = attr.height;
	np->colorMap = attr.colormap;
	np->statusFlags = STAT_NONE;
	np->layer = LAYER_NORMAL;
	np->commands = NULL;
	np->commandCount = 0;

	np->borderFlags = BORDER_OUTLINE | BORDER_TITLE
		| BORDER_MIN | BORDER_MAX | BORDER_CLOSE
		| BORDER_MOVE | BORDER_RESIZE;
	np->borderAction = BA_NONE;

	ReadMotifHints(np);

	ReadClientProtocols(np);

	wmhints = XGetWMHints(display, np->window);
	if(wmhints) {
		if((wmhints->flags & StateHint)
			&& wmhints->initial_state == IconicState) {
			startMapped = 0;
		} else {
			startMapped = 1;
		}
		if((wmhints->flags & WindowGroupHint)) {
			np->groupLeader = wmhints->window_group;
		}
		XFree(wmhints);
	} else {
		startMapped = 1;
	}

	if(startMapped) {
		np->statusFlags |= STAT_MAPPED;
	}
	SetDefaultCursor(np->window);
	Reparent(np);
	PlaceWindow(np, alreadyMapped);

	if(startMapped) {
		XMapWindow(display, np->window);
		XMapWindow(display, np->parent);
	}

	DrawBorder(np);

	AddClientToTray(np);

	RaiseClient(np);

	++clientCount;

	if(np->statusFlags & STAT_STICKY) {
		lvalue = 0xFFFFFFFF;
	} else {
		lvalue = np->desktop;
	}
	XChangeProperty(display, np->window, protNetWMDesktop,
		XA_CARDINAL, 32, PropModeReplace,
		(unsigned char*)&lvalue, 1);

	UpdateState(np);
	SendConfigureEvent(np);

	if(np->desktop != currentDesktop
		&& !(np->statusFlags & STAT_STICKY)) {
		HideClient(np);
	}

}

/****************************************************************************
 ****************************************************************************/
void PlaceWindow(ClientNode *np, int alreadyMapped) {
	int north, west;

	Assert(np);

	GetBorderOffsets(np, &north, &west);

	if(!(np->sizeFlags & PPosition) && !alreadyMapped) {
		np->x = west + rootWidth / 2 - np->width / 2;
		np->y = north + rootHeight / 2 - np->height / 2;
	} else {
		Gravitate(np, 0);
	}

	XMoveWindow(display, np->parent, np->x - west, np->y - north);

}

/****************************************************************************
 ****************************************************************************/
void GetBorderOffsets(ClientNode *np, int *north, int *west) {

	Assert(np);
	Assert(north);
	Assert(west);

	*north = 0;
	*west = 0;

	if(np->borderFlags & BORDER_OUTLINE) {
		*west = BORDER_SIZE;
		if(np->borderFlags & BORDER_TITLE) {
			*north = TITLE_SIZE;
		} else {
			*north = BORDER_SIZE;
		}
	}

}

/****************************************************************************
 ****************************************************************************/
void Gravitate(ClientNode *np, int negate) {
	int north, south, west;
	int northDelta, westDelta;

	Assert(np);

	GetBorderOffsets(np, &north, &west);
	if(north) {
		south = BORDER_SIZE;
	} else {
		south = 0;
	}

	northDelta = 0;
	westDelta = 0;
	switch(np->gravity) {
	case 1: /* NorthWest */
		northDelta = -south;
		westDelta = -west;
		break;
	case 2: /* North */
		northDelta = -south;
		break;
	case 3: /* NorthEast */
		northDelta = -south;
		westDelta = west;
		break;
	case 4: /* West */
		westDelta = -west;
		break;
	case 5: /* Center */
		northDelta = (north + south) / 2;
		westDelta = west;
		break;
	case 6: /* East */
		westDelta = west;
		break;
	case 7: /* SouthWest */
		northDelta = north;
		westDelta = -west;
		break;
	case 8: /* South */
		northDelta = north;
		break;
	case 9: /* SouthEast */
		northDelta = north;
		westDelta = west;
		break;
	default: /* Static */
		break;
	}

	if(negate) {
		np->x += westDelta;
		np->y += northDelta;
	} else {
		np->x -= westDelta;
		np->y -= northDelta;
	}

}

/****************************************************************************
 ****************************************************************************/
void MinimizeClient(ClientNode *np) {

	Assert(np);

	MinimizeTransients(np);
	DrawTray();

}

/****************************************************************************
 ****************************************************************************/
void MinimizeTransients(ClientNode *np) {
	ClientNode *tp;

	Assert(np);

	if(activeClient == np) {
		activeClient = NULL;
		np->statusFlags &= ~STAT_ACTIVE;
	}
	np->statusFlags &= ~STAT_MAPPED;
	XUnmapWindow(display, np->window);
	XUnmapWindow(display, np->parent);
	UpdateState(np);

	for(tp = nodes; tp; tp = tp->next) {
		if(tp->owner == np->window && (tp->statusFlags & STAT_MAPPED)) {
			MinimizeTransients(tp);
		}
	}
}

/****************************************************************************
 ****************************************************************************/
void RestoreClient(ClientNode *np) {
	ClientNode *tp;

	Assert(np);

	XMapWindow(display, np->window);
	XMapWindow(display, np->parent);
	np->statusFlags |= STAT_MAPPED;
	RaiseClient(np);
	UpdateState(np);

	for(tp = nodes; tp; tp = tp->next) {
		if(tp->owner == np->window && !(tp->statusFlags & STAT_MAPPED)) {
			RestoreClient(tp);
		}
	}
}

/****************************************************************************
 ****************************************************************************/
void SetClientLayer(ClientNode *np, int layer) {
	ClientNode *tp;
	long lvalue;

	Assert(np);

	if(layer < LAYER_BOTTOM || layer > LAYER_TOP) {
		return;
	}

	if(np->layer != layer) {

		for(tp = nodes; tp; tp = tp->next) {
			if(tp == np || tp->owner == np->window) {
				tp->layer = layer;
				lvalue = layer;
				XChangeProperty(display, tp->window, protWinLayer,
					XA_CARDINAL, 32, PropModeReplace,
					(unsigned char*)&lvalue, 1);
			}
		}

		RestackClients();

	}

}

/****************************************************************************
 * Set a client's sticky status. This will update transients.
 ****************************************************************************/
void SetClientSticky(ClientNode *np, int isSticky) {
	ClientNode *tp;
	long lvalue;
	int old;

	Assert(np);

	if(np->statusFlags & STAT_STICKY) {
		old = 1;
	} else {
		old = 0;
	}

	if(isSticky && !old) {
		SetClientDesktop(np, currentDesktop);
		for(tp = nodes; tp; tp = tp->next) {
			if(tp == np || tp->owner == np->window) {
				tp->statusFlags |= STAT_STICKY;
				lvalue = 0xFFFFFFFF;
				XChangeProperty(display, tp->window, protNetWMDesktop,
					XA_CARDINAL, 32, PropModeReplace,
					(unsigned char*)&lvalue, 1);
			}
		}
	} else if(!isSticky && old) {
		for(tp = nodes; tp; tp = tp->next) {
			if(tp == np || tp->owner == np->window) {
				tp->statusFlags &= ~STAT_STICKY;
			}
		}
		SetClientDesktop(np, currentDesktop);
	}

}

/****************************************************************************
 * Set a client's desktop. This will update transients.
 ****************************************************************************/
void SetClientDesktop(ClientNode *np, int desktop) {
	ClientNode *tp;
	long lvalue;

	Assert(np);

	if(desktop < 0 || desktop >= desktopCount) {
		return;
	}

	if(np->desktop != desktop && !(np->statusFlags & STAT_STICKY)) {
		for(tp = nodes; tp; tp = tp->next) {
			if(tp == np || tp->owner == np->window) {

				tp->desktop = desktop;

				if(desktop == currentDesktop) {
					ShowClient(tp);
				} else {
					HideClient(tp);
				}

				lvalue = tp->desktop;
				XChangeProperty(display, tp->window, protNetWMDesktop,
					XA_CARDINAL, 32, PropModeReplace,
					(unsigned char*)&lvalue, 1);
			}
		}
		RefreshTray();
	}

}

/****************************************************************************
 * Hide a client without unmapping. This will NOT update transients.
 ****************************************************************************/
void HideClient(ClientNode *np) {

	Assert(np);

	if(activeClient == np) {
		activeClient = NULL;
	}
	np->statusFlags |= STAT_HIDDEN;
	if(np->statusFlags & STAT_MAPPED) {
		XUnmapWindow(display, np->parent);
		UpdateState(np);
	}
}

/****************************************************************************
 * Show a hidden client. This will NOT update transients.
 ****************************************************************************/
void ShowClient(ClientNode *np) {

	Assert(np);

	if(np->statusFlags & STAT_HIDDEN) {
		np->statusFlags &= ~STAT_HIDDEN;
		if(np->statusFlags & STAT_MAPPED) {
			XMapWindow(display, np->parent);
			if(np->statusFlags & STAT_ACTIVE) {
				FocusClient(np);
			}
		}
	}
}

/****************************************************************************
 ****************************************************************************/
void MaximizeClient(ClientNode *np) {
	int north, west;

	Assert(np);

	if(np->borderFlags & BORDER_OUTLINE) {
		west = BORDER_SIZE;
	} else {
		west = 0;
	}
	if(np->borderFlags & BORDER_TITLE) {
		north = TITLE_SIZE;
	} else {
		north = west;
	}

	if(np->statusFlags & STAT_MAXIMIZED) {
		np->x = np->oldx;
		np->y = np->oldy;
		np->width = np->oldWidth;
		np->height = np->oldHeight;
		np->statusFlags &= ~STAT_MAXIMIZED;
	} else {
		np->oldx = np->x;
		np->oldy = np->y;
		np->oldWidth = np->width;
		np->oldHeight = np->height;

		np->x = west;
		np->y = north;
		np->width = rootWidth - 2 * BORDER_SIZE;
		np->height = rootHeight - TITLE_SIZE - BORDER_SIZE - TRAY_HEIGHT;

		if(np->width > np->maxWidth) {
			np->width = np->maxWidth;
		}
		if(np->height > np->maxHeight) {
			np->height = np->maxHeight;
		}

		np->width -= np->width % np->xinc;
		np->height -= np->height % np->yinc;

		np->statusFlags |= STAT_MAXIMIZED;
	}

	XMoveResizeWindow(display, np->parent,
		np->x - west, np->y - north,
		np->width + 2 * west,
		np->height + north + west);
	XMoveResizeWindow(display, np->window, west,
		north, np->width, np->height);
	SendConfigureEvent(np);
}

/****************************************************************************
 ****************************************************************************/
void FocusClient(ClientNode *np) {

	Assert(np);

	if(!(np->statusFlags & STAT_MAPPED)) {
		return;
	}

	if(activeClient && activeClient != np) {
		activeClient->statusFlags &= ~STAT_ACTIVE;
		DrawBorder(activeClient);
	}
	np->statusFlags |= STAT_ACTIVE;
	activeClient = np;
	if(np->colorMap) {
		XInstallColormap(display, np->colorMap);
	}

	if(np->protocols & PROT_TAKE_FOCUS) {
		SendClientMessage(np, protWMProtocols, protWMTakeFocus);
	}

	DrawBorder(np);
	DrawTray();

	XSetInputFocus(display, np->window, RevertToPointerRoot, CurrentTime);

	XChangeProperty(display, rootWindow, protNetActiveWindow,
		XA_WINDOW, 32, PropModeReplace,
		(unsigned char*)&np->window, 1);

}

/****************************************************************************
 ****************************************************************************/
void RefocusClient() {
	if(activeClient) {
		FocusClient(activeClient);
	}
}

/****************************************************************************
 ****************************************************************************/
void DeleteClient(ClientNode *np) {
	if(np->protocols & PROT_DELETE) {
		SendClientMessage(np, protWMProtocols, protWMDeleteWindow);
	} else {
		XKillClient(display, np->window);
	}
}

/****************************************************************************
 ****************************************************************************/
void RaiseClient(ClientNode *np) {
	ClientNode *tp, *next;

	Assert(np);
	Assert(nodes);
	Assert(nodeTail);

	if(np->layer > LAYER_BOTTOM) {

		/* Raise the window */
		if(nodes != np) {
			Assert(np->prev);
			np->prev->next = np->next;
			if(np->next) {
				np->next->prev = np->prev;
			} else {
				nodeTail = np->prev;
			}
			np->next = nodes;
			nodes->prev = np;
			np->prev = NULL;
			nodes = np;
		}

		Assert(nodes);

		/* Place any transient windows on top of the owner */
		for(tp = nodes->next; tp; tp = tp->next) {
			if(tp->owner == np->window) {
				next = tp->next;
				Assert(tp->prev);
				tp->prev->next = tp->next;
				if(tp->next) {
					tp->next->prev = tp->prev;
				} else {
					nodeTail = tp->prev;
				}
				tp->next = nodes;
				nodes->prev = tp;
				tp->prev = NULL;
				nodes = tp;
				tp = next;
			}
			if(!tp) {
				break;
			}
		}

		Assert(nodes);
		Assert(nodeTail);

		RestackClients();
	}
}

/****************************************************************************
 ****************************************************************************/
void RestackClients() {
	ClientNode *np;
	int layer, index;

	Assert(stack);

	if(clientCount + 1 >= stackSize) {
		do {
			stackSize += STACK_BLOCK_SIZE;
		} while(clientCount + 1 >= stackSize);
		stack = Reallocate(stack, sizeof(Window) * stackSize);
	}

	/* Prepare the stacking array. */
	index = 0;
	for(layer = LAYER_TOP; layer >= LAYER_BOTTOM; layer--) {
		for(np = nodes; np; np = np->next) {
			if(np->layer == layer && (np->statusFlags & STAT_MAPPED)
				&& !(np->statusFlags & STAT_HIDDEN)) {
				stack[index++] = np->parent;
			}
		}
		if(layer == LAYER_TRAY) {
			stack[index++] = trayWindow;
		}
	}

	XRestackWindows(display, stack, index);

}

/****************************************************************************
 ****************************************************************************/
void SendClientMessage(ClientNode *np, Atom type, Atom message) {
	XEvent event;
	int status;

	Assert(np);

	memset(&event, 0, sizeof(event));
	event.xclient.type = ClientMessage;
	event.xclient.window = np->window;
	event.xclient.message_type = type;
	event.xclient.format = 32;
	event.xclient.data.l[0] = message;
	event.xclient.data.l[1] = CurrentTime;

	status = XSendEvent(display, np->window, False, 0, &event);
	if(status == False) {
		Debug("SendClientMessage failed");
	}
}

/****************************************************************************
 ****************************************************************************/
#ifdef USE_SHAPE
void SetShape(ClientNode *np) {
	XRectangle rect[4];
	int north, west;

	Assert(np);

	if(np->borderFlags & BORDER_OUTLINE) {
		west = BORDER_SIZE;
	} else {
		west = 0;
	}
	if(np->borderFlags & BORDER_TITLE) {
		north = TITLE_SIZE;
	} else {
		north = west;
	}

	XShapeCombineShape(display, np->parent, ShapeBounding, west, north,
		np->window, ShapeBounding, ShapeSet);

	if(north > 0) {

		/* Top */
		rect[0].x = 0;
		rect[0].y = 0;
		rect[0].width = np->width + west * 2;
		rect[0].height = north;

		/* Left */
		rect[1].x = 0;
		rect[1].y = 0;
		rect[1].width = west;
		rect[1].height = np->height + west + north;

		/* Right */
		rect[2].x = np->width + west;
		rect[2].y = 0;
		rect[2].width = west;
		rect[2].height = np->height + west + north;

		/* Bottom */
		rect[3].x = 0;
		rect[3].y = np->height + north;
		rect[3].width = np->width + west * 2;
		rect[3].height = west;

		XShapeCombineRectangles(display, np->parent, ShapeBounding,
			0, 0, rect, 4, ShapeUnion, Unsorted);

	}
}
#endif /* USE_SHAPE */

/****************************************************************************
 ****************************************************************************/
void MoveClient(ClientNode *np) {
	XEvent event;
	int startx, starty;
	int oldx, oldy;
	int doMove = 0;

	Assert(np);

	if(!(np->borderFlags & BORDER_OUTLINE)
		|| !(np->borderFlags & BORDER_MOVE)
		|| (np->statusFlags & STAT_MAXIMIZED)) {
		return;
	}

	if(!GrabMouseForMove()) {
		return;
	}

	GetMousePosition(&startx, &starty);
	oldx = np->x;
	oldy = np->y;

	for(;;) {
		XMaskEvent(display, ButtonReleaseMask | PointerMotionMask
			| ExposureMask, &event);

		if(!(GetMouseMask() & (Button1Mask | Button2Mask))) {
			goto StopMove;
		}

		switch(event.type) {
		case Expose:
			HandleExpose(&event.xexpose);
			break;
		case MotionNotify:

			while(XCheckTypedEvent(display, MotionNotify, &event));

			np->x = oldx + event.xmotion.x - startx;
			np->y = oldy + event.xmotion.y - starty;

			if(!doMove && (abs(np->x - oldx) > MOVE_DELTA
				|| abs(np->y - oldy) > MOVE_DELTA)) {
				CreateMoveWindow();
				doMove = 1;
			}

			if(doMove) {
				if(np->borderFlags & BORDER_TITLE) {
					XMoveWindow(display, np->parent, np->x - BORDER_SIZE,
						np->y - TITLE_SIZE);
				} else {
					XMoveWindow(display, np->parent, np->x - BORDER_SIZE,
						np->y - BORDER_SIZE);
				}
				UpdateMoveWindow(np->x, np->y);
				UpdatePager();
			}

			break;
		default:
			break;
		}
	}

StopMove:
	XUngrabPointer(display, CurrentTime);

	if(!doMove) {
		np->x = oldx;
		np->y = oldy;
	} else {
		DestroyMoveWindow();
	}

	if(np->borderFlags & BORDER_TITLE) {
		XMoveWindow(display, np->parent, np->x - BORDER_SIZE,
			np->y - TITLE_SIZE);
	} else {
		XMoveWindow(display, np->parent, np->x - BORDER_SIZE,
			np->y - BORDER_SIZE);
	}
	SendConfigureEvent(np);

}

/****************************************************************************
 ****************************************************************************/
void DrawOutline(ClientNode *np) {
	int north;

	Assert(np);

	if(np->borderFlags & BORDER_TITLE) {
		north = TITLE_SIZE;
	} else {
		north = BORDER_SIZE;
	}

	XDrawRectangle(display, rootWindow, invertedGC,
		np->x - BORDER_SIZE, np->y - north,
		np->width + BORDER_SIZE + BORDER_SIZE,
		np->height + north + BORDER_SIZE);
}

/****************************************************************************
 ****************************************************************************/
void ResizeClient(ClientNode *np, BorderActionType action) {

	XEvent event;
	int startx, starty;
	int oldx, oldy;
	int oldw, oldh;
	int gwidth, gheight;
	int delta;
	int north;

	Assert(np);

	if(!(np->borderFlags & BORDER_OUTLINE)
		|| !(np->borderFlags & BORDER_RESIZE)
		|| (np->statusFlags & STAT_MAXIMIZED)) {
		return;
	}

	if(!GrabMouseForResize(action)) {
		return;
	}

	GetMousePosition(&startx, &starty);

	DrawOutline(np);

	oldx = np->x;
	oldy = np->y;
	oldw = np->width;
	oldh = np->height;

	gwidth = (np->width - np->baseWidth) / np->xinc;
	gheight = (np->height - np->baseHeight) / np->yinc;

	CreateResizeWindow();
	UpdateResizeWindow(gwidth, gheight);

	for(;;) {

		XMaskEvent(display, ButtonReleaseMask | PointerMotionMask, &event);

		if(!(GetMouseMask() & Button1Mask)) {
			goto StopResize;
		}

		switch(event.type) {
		case MotionNotify:

			DrawOutline(np);

			if(action & BA_RESIZE_N) {
				delta = (event.xmotion.y - starty) / np->yinc;
				delta *= np->yinc;
				if(oldh - delta >= np->minHeight
					&& oldh - delta <= np->maxHeight) {
					np->height = oldh - delta;
					np->y = oldy + delta;
				}
			}
			if(action & BA_RESIZE_S) {
				delta = (event.xmotion.y - starty) / np->yinc;
				delta *= np->yinc;
				np->height = oldh + delta;
				np->height = Max(np->height, np->minHeight);
				np->height = Min(np->height, np->maxHeight);
			}
			if(action & BA_RESIZE_E) {
				delta = (event.xmotion.x - startx) / np->xinc;
				delta *= np->xinc;
				np->width = oldw + delta;
				np->width = Max(np->width, np->minWidth);
				np->width = Min(np->width, np->maxWidth);
			}
			if(action & BA_RESIZE_W) {
				delta = (event.xmotion.x - startx) / np->xinc;
				delta *= np->xinc;
				if(oldw - delta >= np->minWidth
					&& oldw - delta <= np->maxWidth) {
					np->width = oldw - delta;
					np->x = oldx + delta;
				}
			}

			gwidth = (np->width - np->baseWidth) / np->xinc;
			gheight = (np->height - np->baseHeight) / np->yinc;

			DrawOutline(np);
			UpdateResizeWindow(gwidth, gheight);
			UpdatePager();

			break;
		default:
			break;
		}
	}
StopResize:
	DrawOutline(np);
	XUngrabPointer(display, CurrentTime);

	DestroyResizeWindow();

	if(np->borderFlags & BORDER_TITLE) {
		north = TITLE_SIZE;
	} else {
		north = BORDER_SIZE;
	}

	XMoveResizeWindow(display, np->parent,
		np->x - BORDER_SIZE, np->y - north,
		np->width + BORDER_SIZE + BORDER_SIZE,
		np->height + north + BORDER_SIZE);
	XMoveResizeWindow(display, np->window, BORDER_SIZE,
		north, np->width, np->height);

	SendConfigureEvent(np);

}

/****************************************************************************
 ****************************************************************************/
void RemoveClient(ClientNode *np) {
	Window w;

	Assert(np);
	Assert(nodes);
	Assert(nodeTail);

	if(activeClient == np) {

		w = None;
		XChangeProperty(display, rootWindow, protNetActiveWindow,
			XA_WINDOW, 32, PropModeReplace,
			(unsigned char*)&w, 1);

		activeClient = NULL;

	}

	np->statusFlags &= ~STAT_MAPPED;

	if(shouldExit) {
		Gravitate(np, 1);
		XRemoveFromSaveSet(display, np->window);
		if(!(np->statusFlags & STAT_MAPPED)) {
			XMapWindow(display, np->window);
		}
		XUngrabButton(display, AnyButton, AnyModifier, np->window);
		XReparentWindow(display, np->window, rootWindow, np->x, np->y);
	}
	if(np->parent) {
		XFreeGC(display, np->parentGC);
		XDestroyWindow(display, np->parent);
	}

	if(np->name) {
		XFree(np->name);
	}
	if(np->colorMap) {
		XUninstallColormap(display, np->colorMap);
	}
	if(np->commands) {
		XFreeStringList(np->commands);
	}

	if(np->next) {
		np->next->prev = np->prev;
	}

	if(np->prev) {
		np->prev->next = np->next;
	} else {
		nodes = np->next;
	}
	if(!np->next) {
		nodeTail = np->prev;
	}

	Release(np);

	RemoveClientFromTray(np);

	--clientCount;

}

/****************************************************************************
 ****************************************************************************/
ClientNode *FindClientByWindow(Window w) {
	ClientNode *np;

	for(np = nodes; np; np = np->next) {
		if(np->window == w || np->parent == w) {
			return np;
		}
	}
	return NULL;
}

/****************************************************************************
 ****************************************************************************/
ClientNode *FindClientByParent(Window p) {
	ClientNode *np;

	for(np = nodes; np; np = np->next) {
		if(np->parent == p) {
			return np;
		}
	}
	return NULL;
}

/****************************************************************************
 ****************************************************************************/
void Reparent(ClientNode *np) {
	XSetWindowAttributes attr;
	int x, y, width, height;

	Assert(np);

	XAddToSaveSet(display, np->window);
	XSelectInput(display, np->window, PropertyChangeMask
		| EnterWindowMask | ColormapChangeMask);

	XGrabButton(display, AnyButton, AnyModifier, np->window,
		True, ButtonPressMask, GrabModeSync, GrabModeAsync, None, None);

	XGrabKey(display, XKeysymToKeycode(display, XK_Tab), altMask,
		np->window, True, GrabModeSync, GrabModeAsync);

	attr.override_redirect = True;
	attr.event_mask = ButtonPressMask | ExposureMask | PointerMotionMask
		| SubstructureRedirectMask | SubstructureNotifyMask
		| EnterWindowMask | KeyPressMask;
	attr.background_pixel = colors[COLOR_BORDER];

	if(np->borderFlags & BORDER_OUTLINE) {
		x = np->x - BORDER_SIZE;
		width = np->width + BORDER_SIZE + BORDER_SIZE;
		if(np->borderFlags & BORDER_TITLE) {
			y = np->y - TITLE_SIZE;
			height = np->height + TITLE_SIZE + BORDER_SIZE;
		} else {
			y = np->y - BORDER_SIZE;
			height = np->height + 2 * BORDER_SIZE;
		}
	} else {
		x = np->x;
		y = np->y;
		width = np->width;
		height = np->height;
	}

	np->parent = XCreateWindow(display, rootWindow,
		x, y, width, height, 0,
		DefaultDepth(display, 0), InputOutput,
		DefaultVisual(display, 0),
		CWOverrideRedirect | CWBackPixel | CWEventMask, &attr);
	np->parentGC = XCreateGC(display, np->parent, 0, NULL);

#ifdef USE_SHAPE
	if(haveShape) {
		XShapeSelectInput(display, np->window, ShapeNotifyMask);
		CheckShape(np);
	}
#endif

	XSetWindowBorderWidth(display, np->window, 0);
	if(np->borderFlags & BORDER_OUTLINE) {
		if(np->borderFlags & BORDER_TITLE) {
			XReparentWindow(display, np->window, np->parent, BORDER_SIZE,
				TITLE_SIZE);
		} else {
			XReparentWindow(display, np->window, np->parent, BORDER_SIZE,
				BORDER_SIZE);
		}
	} else {
		XReparentWindow(display, np->window, np->parent, 0, 0);
	}

	SendConfigureEvent(np);
}

/****************************************************************************
 ****************************************************************************/
#ifdef USE_SHAPE
void CheckShape(ClientNode *np) {
	int xb, yb;
	int xc, yc;
	unsigned int wb, hb;
	unsigned int wc, hc;
	Bool boundingShaped, clipShaped;

	XShapeQueryExtents(display, np->window, &boundingShaped,
		&xb, &yb, &wb, &hb, &clipShaped, &xc, &yc, &wc, &hc);

	if(boundingShaped == True) {
		SetShape(np);
	}
}
#endif

/****************************************************************************
 ****************************************************************************/
void SendConfigureEvent(ClientNode *np) {
	XConfigureEvent event;

	Assert(np);

	event.type = ConfigureNotify;
	event.event = np->window;
	event.window = np->window;
	event.x = np->x;
	event.y = np->y;
	event.width = np->width;
	event.height = np->height;
	event.border_width = 0;
	event.above = None;
	event.override_redirect = False;

	XSendEvent(display, np->window, False, StructureNotifyMask,
		(XEvent*)&event);
}

/****************************************************************************
 ****************************************************************************/
void ReleaseClients() {

	while(nodes) {
		RemoveClient(nodes);
	}
	Release(stack);

}

/****************************************************************************
 ****************************************************************************/
void UpdateState(ClientNode *np) {
	long data[2];

	Assert(np);

	if(np->statusFlags & STAT_MAPPED) {
		data[0] = NormalState;
	} else {
		data[0] = IconicState;
	}
	data[1] = None;

	XChangeProperty(display, np->window, protWMState, protWMState, 32,
		PropModeReplace, (unsigned char*)data, 2);

}

/****************************************************************************
 ****************************************************************************/
#ifdef USE_MWM_HINTS
PropMwmHints *GetMwmHints(Window w) {
	Atom type;
	unsigned long itemCount, bytesLeft;
	unsigned char *data;
	int format;

	if(XGetWindowProperty(display, w, mwmHints, 0L, 20L, False,
		mwmHints, &type, &format, &itemCount, &bytesLeft, &data)
		== Success) {
		if(itemCount >= PROP_MOTIF_WM_HINTS_ELEMENTS) {
			return (PropMwmHints*)data;
		}
	}
	return NULL;
}
#endif

/****************************************************************************
 * Read Motif hints.
 ****************************************************************************/
void ReadMotifHints(ClientNode *np) {
#ifdef USE_MWM_HINTS
	PropMwmHints *mhints;

	Assert(np);

	mhints = GetMwmHints(np->window);
	if(mhints) {

		if((mhints->flags & MWM_HINTS_FUNCTIONS)
			&& !(mhints->functions & MWM_FUNC_ALL)) {

			if(!(mhints->functions & MWM_FUNC_RESIZE)) {
				np->borderFlags &= ~BORDER_RESIZE;
			}
			if(!(mhints->functions & MWM_FUNC_MOVE)) {
				np->borderFlags &= ~BORDER_MOVE;
			}
			if(!(mhints->functions & MWM_FUNC_MINIMIZE)) {
				np->borderFlags &= ~BORDER_MIN;
			}
			if(!(mhints->functions & MWM_FUNC_MAXIMIZE)) {
				np->borderFlags &= ~BORDER_MAX;
			}
			if(!(mhints->functions & MWM_FUNC_CLOSE)) {
				np->borderFlags &= ~BORDER_CLOSE;
			}
		}

		if((mhints->flags & MWM_HINTS_DECORATIONS)
			&& !(mhints->decorations & MWM_DECOR_ALL)) {

			if(!(mhints->decorations & MWM_DECOR_BORDER)) {
				np->borderFlags &= ~BORDER_OUTLINE;
			}
			if(!(mhints->decorations & MWM_DECOR_TITLE)) {
				np->borderFlags &= ~BORDER_TITLE;
			}
			if(!(mhints->decorations & MWM_DECOR_MINIMIZE)) {
				np->borderFlags &= ~BORDER_MIN;
			}
			if(!(mhints->decorations & MWM_DECOR_MAXIMIZE)) {
				np->borderFlags &= ~BORDER_MAX;
			}
		}

		XFree(mhints);
	}
#endif
}

