/****************************************************************************
 ****************************************************************************/

#include "jwm.h"

static void DispatchBorderButtonEvent(XButtonEvent *event, ClientNode *np);

/****************************************************************************
 ****************************************************************************/
void HandleButtonPress(XButtonEvent *event) {
	int x, y;
	ClientNode *np;

	np = FindClientByParent(event->window);
	if(np) {
		RaiseClient(np);
		switch(event->button) {
		case Button1:
			DispatchBorderButtonEvent(event, np);
			break;
		case Button2:
			MoveClient(np);
			break;
		case Button3:
			x = event->x + np->x;
			y = event->y + np->y;
			if(np->borderFlags & BORDER_OUTLINE) {
				x -= BORDER_SIZE;
				if(np->borderFlags & BORDER_TITLE) {
					y -= TITLE_SIZE;
				} else {
					y -= BORDER_SIZE;
				}
			} else {
				if(np->borderFlags & BORDER_TITLE) {
					y -= TITLE_SIZE;
				}
			}
			ShowWindowMenu(np, x, y);
			break;
		default:
			break;
		}
	} else if(event->window == trayWindow) {
		HandleTrayClick(event);
	} else if(event->window == rootWindow) {
		ShowRootMenu(event->x, event->y);
	} else {
		np = FindClientByWindow(event->window);
		if(np) {
			RaiseClient(np);
			XAllowEvents(display, ReplayPointer, CurrentTime);
		}
	}

	UpdatePager();
}

/****************************************************************************
 ****************************************************************************/
void HandleKeyPress(XKeyEvent *event) {
	KeyCode code;

	if(event->state == altMask) {
		code = XKeysymToKeycode(display, XK_Tab);
		if(event->keycode == code) {
			FocusNext();
		}
	}

}

/****************************************************************************
 ****************************************************************************/
void HandleConfigureRequest(XConfigureRequestEvent *event) {
	XWindowChanges wc;
	ClientNode *np;
	int north, south, east, west;
	int changed;

	np = FindClientByWindow(event->window);
	if(np) {

		changed = 0;
		if(event->value_mask & CWWidth && event->width != np->width) {
			np->width = event->width;
			changed = 1;
		}
		if(event->value_mask & CWHeight && event->height != np->height) {
			np->height = event->height;
			changed = 1;
		}
		if(event->value_mask & CWX && event->x != np->x) {
			np->x = event->x;
			changed = 1;
		}
		if(event->value_mask & CWY && event->y != np->y) {
			np->y = event->y;
			changed = 1;
		}

		if(!changed) {
			return;
		}

		if(np->borderFlags & BORDER_OUTLINE) {
			east = BORDER_SIZE;
			west = BORDER_SIZE;
			south = BORDER_SIZE;
			if(np->borderFlags & BORDER_TITLE) {
				north = TITLE_SIZE;
			} else {
				north = BORDER_SIZE;
			}
		} else {
			north = 0;
			south = 0;
			east = 0;
			west = 0;
		}

		wc.stack_mode = event->detail;
		wc.sibling = event->above;
		wc.border_width = 0;

		wc.x = west;
		wc.y = north;
		wc.width = np->width;
		wc.height = np->height;
		XConfigureWindow(display, np->window, event->value_mask, &wc);

		wc.x = np->x;
		wc.y = np->y;
		wc.width = np->width + east + west;
		wc.height = np->height + north + south;
		wc.border_width = 1;
		XConfigureWindow(display, np->parent, event->value_mask, &wc);

	} else {
		wc.stack_mode = event->detail;
		wc.sibling = event->above;
		wc.border_width = event->border_width;
		wc.x = event->x;
		wc.y = event->y;
		wc.width = event->width;
		wc.height = event->height;
		XConfigureWindow(display, event->window, event->value_mask, &wc);
	}

}

/****************************************************************************
 ****************************************************************************/
void HandleEnterNotify(XCrossingEvent *event) {
	ClientNode *np;

	np = FindClientByWindow(event->window);
	if(np && (np->statusFlags & STAT_MAPPED)
		&& !(np->statusFlags & STAT_ACTIVE)) {
		FocusClient(np);
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleExpose(XExposeEvent *event) {
	ClientNode *np;

	np = FindClientByWindow(event->window);
	if(np) {
		DrawBorder(np);
	} else if(event->window == trayWindow) {
		DrawTray();
	}
}

/****************************************************************************
 ****************************************************************************/
void HandlePropertyNotify(XPropertyEvent *event) {
	ClientNode *np;
	char *atomName;

	np = FindClientByWindow(event->window);
	if(np) {
		switch(event->atom) {
		case XA_WM_NAME:
			ReadWMName(np);
			break;
		case XA_WM_NORMAL_HINTS:
			ReadWMNormalHints(np);
			break;
		case XA_WM_HINTS:
			break;
		case XA_WM_ICON_NAME:
		case XA_WM_CLIENT_MACHINE:
			break;
		default:
			if(event->atom == protWMState) {
			} else if(event->atom == protWMLocaleName) {
			} else if(event->atom == protNetWMName) {
			} else if(event->atom == protNetWMIcon) {
			} else if(event->atom == protWMCommand) {
				ReadWMCommand(np);
			} else if(event->atom == protNetWMDesktop) {
				ReadNetWMDesktop(np);
			} else if(event->atom == protWinLayer) {
				ReadWinLayer(np);
			} else {
				atomName = XGetAtomName(display, event->atom);
				Debug("Uknown PropertyNotify: %s", atomName);
				XFree(atomName);
			}
			break;
		}

		DrawBorder(np);
		DrawTray();
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleClientMessage(XClientMessageEvent *event) {
	ClientNode *np;
	long mask, flags;
	char *atomName;

	np = FindClientByWindow(event->window);
	if(np) {
		if(event->message_type == protWinState) {

			mask = event->data.l[0];
			flags = event->data.l[1];

			if(mask & WIN_STATE_STICKY) {
				if(flags & WIN_STATE_STICKY) {
					SetClientSticky(np, 1);
				} else {
					SetClientSticky(np, 0);
				}
			}

		} else if(event->message_type == protWinLayer) {

			np->layer = event->data.l[0];
			SetClientLayer(np, event->data.l[0]);

		} else if(event->message_type == protWMChangeState) {

			switch(event->data.l[0]) {
			case IconicState:
				MinimizeClient(np);
				break;
			case NormalState:
				RestoreClient(np);
				break;
			default:
				break;
			}

		} else if(event->message_type == protNetActiveWindow) {

			if(event->data.l[0] == 0) {
				RaiseClient(np);
			}

		} else if(event->message_type == protNetWMDesktop) {

			if(event->data.l[0] == 0xFFFFFFFF) {
				SetClientSticky(np, 1);
			} else {
				if(event->data.l[0] >= 0 && event->data.l[0] < desktopCount) {
					np->statusFlags &= ~STAT_STICKY;
					SetClientDesktop(np, event->data.l[0]);
				}
			}

		} else {
			atomName = XGetAtomName(display, event->message_type);
			Debug("Uknown ClientMessage: %s", atomName);
			XFree(atomName);
		}
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleMotionNotify(XMotionEvent *event) {
	ClientNode *np;
	Cursor cur;
	BorderActionType action;

	np = FindClientByParent(event->window);
	if(np && (np->borderFlags & BORDER_OUTLINE)) {
		action = GetBorderActionType(np, event->x, event->y);
		if(np->borderAction != action) {
			np->borderAction = action;
			cur = GetFrameCursor(action);
			XDefineCursor(display, np->parent, cur);
		}
	}

}
/****************************************************************************
 ****************************************************************************/
void HandleShapeEvent(XShapeEvent *event) {
	ClientNode *np;

	np = FindClientByWindow(event->window);
	if(np) {
		SetShape(np);
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleColormapChange(XColormapEvent *event) {
	ClientNode *np;

	if(event->new) {
		np = FindClientByWindow(event->window);
		if(np) {
			np->colorMap = event->colormap;
			XInstallColormap(display, np->colorMap);
		}
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleMapRequest(XMapEvent *event) {
	ClientNode *np;

	Assert(event);

	np = FindClientByWindow(event->window);
	if(!np) {
		AddClientWindow(event->window, 0);
	} else {
		np->statusFlags |= STAT_MAPPED;
		XMapWindow(display, np->window);
		XMapWindow(display, np->parent);
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleUnmapNotify(XUnmapEvent *event) {
	ClientNode *np;

	Assert(event);

	np = FindClientByWindow(event->window);
	if(np && np->window == event->window) {
		np->statusFlags &= ~STAT_MAPPED;
		XUnmapWindow(display, np->parent);
	}

}

/****************************************************************************
 ****************************************************************************/
void HandleDestroyNotify(XDestroyWindowEvent *event) {
	ClientNode *np;

	np = FindClientByWindow(event->window);
	if(np && np->window == event->window) {
		RemoveClient(np);
	}

}

/****************************************************************************
 ****************************************************************************/
void DispatchBorderButtonEvent(XButtonEvent *event, ClientNode *np) {
	BorderActionType action;

	if(!(np->borderFlags & BORDER_OUTLINE)) {
		return;
	}

	action = GetBorderActionType(np, event->x, event->y);

	switch(action & 0x0F) {
	case BA_RESIZE:
		ResizeClient(np, action);
		break;
	case BA_MOVE:
		MoveClient(np);
		break;
	case BA_CLOSE:
		DeleteClient(np);
		break;
	case BA_MAXIMIZE:
		MaximizeClient(np);
		break;
	case BA_MINIMIZE:
		MinimizeClient(np);
		break;
	default:
		break;
	}

}

