/****************************************************************************
 ****************************************************************************/

#include "jwm.h"

#ifdef USE_SHAPE
static int shapeEvent;
#endif

static char *CONFIG_FILE = "/.jwmrc";

static void Initialize();
static void EventLoop();
static void Shutdown();
static void HandleExit();
static void DoExit(int code);

static char *configPath = NULL;
static char *displayString = NULL;

/****************************************************************************
 ****************************************************************************/
int main(int argc, char *argv[]) {
	char *temp;
	int x;

	StartDebug();

	temp = getenv("HOME");
	if(temp) {
		configPath = Allocate(strlen(temp) + strlen(CONFIG_FILE) + 1);
		strcpy(configPath, temp);
		strcat(configPath, CONFIG_FILE);
	} else {
		configPath = Allocate(strlen(CONFIG_FILE) + 1);
		strcpy(configPath, CONFIG_FILE);
	}

	for(x = 1; x < argc; x++) {
		if(!strcmp(argv[x], "-v")) {
			DisplayAbout();
			DoExit(0);
		} else if(!strcmp(argv[x], "-h")) {
			DisplayHelp();
			DoExit(0);
		} else if(!strcmp(argv[x], "-p")) {
			ParseConfig(configPath);
			DestroyRootMenu();
			ReleaseFonts();
			ReleaseColors();
			DoExit(0);
		} else if(!strcmp(argv[x], "-display") && x + 1 < argc) {
			displayString = argv[++x];
		} else {
			DisplayUsage();
			DoExit(1);
		}
	}

	do {
		shouldExit = 0;
		shouldRestart = 0;
		ParseConfig(configPath);
		Initialize();
		EventLoop();
		Shutdown();
	} while(shouldRestart);
	Release(configPath);
	StopDebug();

	return 0;
}

/****************************************************************************
 ****************************************************************************/
void DoExit(int code) {
	if(configPath) {
		Release(configPath);
	}
	StopDebug();
	exit(code);
}

/****************************************************************************
 ****************************************************************************/
void EventLoop() {
	struct timeval timeout;
	int fd;
	fd_set fds;
	XEvent event;

	fd = XConnectionNumber(display);

	while(!shouldExit) {

		FD_ZERO(&fds);
		FD_SET(fd, &fds);
		timeout.tv_sec = MAIN_DELTA;
		timeout.tv_usec = 0;
		if(select(fd + 1, &fds, NULL, NULL, &timeout) <= 0) {
			UpdateTime();
		}

		while(XPending(display) > 0) {
			XNextEvent(display, &event);
			switch(event.type) {
			case ButtonPress:
				HandleButtonPress(&event.xbutton);
				break;
			case KeyPress:
				HandleKeyPress(&event.xkey);
				break;
			case ConfigureRequest:
				HandleConfigureRequest(&event.xconfigurerequest);
				break;
			case MapRequest:
				HandleMapRequest(&event.xmap);
				break;
			case EnterNotify:
				HandleEnterNotify(&event.xcrossing);
				break;
			case PropertyNotify:
				HandlePropertyNotify(&event.xproperty);
				break;
			case ClientMessage:
				HandleClientMessage(&event.xclient);
				break;
			case UnmapNotify:
				HandleUnmapNotify(&event.xunmap);
				break;
			case Expose:
				HandleExpose(&event.xexpose);
				break;
			case MotionNotify:
				HandleMotionNotify(&event.xmotion);
				break;
			case ColormapNotify:
				HandleColormapChange(&event.xcolormap);
				break;
			case DestroyNotify:
				HandleDestroyNotify(&event.xdestroywindow);
				break;
			case KeyRelease:
			case ButtonRelease:
			case ConfigureNotify:
			case CreateNotify:
			case MapNotify:
			case NoExpose:
			case ReparentNotify:
				break;
			default:
				if(0) {
#ifdef USE_SHAPE
				} else if(haveShape && event.type == shapeEvent) {
					HandleShapeEvent((XShapeEvent*)&event);
#endif
				} else {
					Debug("Unknown event type: %d", event.type);
				}
				break;
			}
		}

	}
}

/****************************************************************************
 ****************************************************************************/
void Initialize() {
	XGCValues gcvalues;
	XSetWindowAttributes attr;
	XModifierKeymap *modmap;
	KeyCode altCode;
	int temp;
	int x;

	initializing = 1;
	display = XOpenDisplay(displayString);
	if(!display) {
		if(displayString) {
			printf("error: could not open display %s\n", displayString);
		} else {
			printf("error: could not open display\n");
		}
		DoExit(1);
	}

	XSetErrorHandler(ErrorHandler);

/*
	wmState = XInternAtom(display, "WM_STATE", False);
	wmChangeState = XInternAtom(display, "WM_CHANGE_STATE", False);
	wmProtocols = XInternAtom(display, "WM_PROTOCOLS", False);
	wmDelete = XInternAtom(display, "WM_DELETE_WINDOW", False);
	wmTakeFocus = XInternAtom(display, "WM_TAKE_FOCUS", False);
	wmColormaps = XInternAtom(display, "WM_COLORMAP_WINDOWS", False);
	wmClientMachine = XInternAtom(display, "WM_CLIENT_MACHINE", False);
	wmLocaleName = XInternAtom(display, "WM_LOCALE_NAME", False);
	wmRunning = XInternAtom(display, "_JWM_RUNNING", False);
*/
#ifdef USE_MWM_HINTS
	mwmHints = XInternAtom(display, _XA_MWM_HINTS, False);
#endif

	InitializeColors();
	InitializeFonts();
	InitializeCursors();

	rootWindow = RootWindow(display, 0);
	rootWidth = DisplayWidth(display, DefaultScreen(display));
	rootHeight = DisplayHeight(display, DefaultScreen(display));

	attr.event_mask = SubstructureRedirectMask | SubstructureNotifyMask
		| PropertyChangeMask | ColormapChangeMask | ButtonPressMask
		| ButtonReleaseMask;
	XChangeWindowAttributes(display, rootWindow, CWEventMask, &attr);

	SetDefaultCursor(rootWindow);

	gcvalues.function = GXinvert;
	gcvalues.subwindow_mode = IncludeInferiors;
	gcvalues.line_width = 2;
	invertedGC = XCreateGC(display, rootWindow, GCFunction
		| GCSubwindowMode | GCLineWidth, &gcvalues);

	signal(SIGTERM, HandleExit);
	signal(SIGINT, HandleExit);
	signal(SIGHUP, HandleExit);

#ifdef USE_SHAPE
	haveShape = XShapeQueryExtension(display, &shapeEvent, &temp);
	if(!haveShape) {
		Debug("No shape extension.");
	}
#endif

	InitializeLoadDisplay();
	CreateTray();
	InitializePager();

	/* Determine the alt key mask. */
	altCode = XKeysymToKeycode(display, XK_Alt_L);
	modmap = XGetModifierMapping(display);

	for(x = 0; x < 8 * modmap->max_keypermod; x++) {
		if(modmap->modifiermap[x] == altCode) {
			altMask = 1 << (x / modmap->max_keypermod);
		}
	}
	XFreeModifiermap(modmap);

	InitializeProtocols();

	LoadClients();

	initializing = 0;

}

/****************************************************************************
 ****************************************************************************/
void HandleExit() {
	signal(SIGTERM, HandleExit);
	signal(SIGINT, HandleExit);
	signal(SIGHUP, HandleExit);
	shouldExit = 1;
}

/****************************************************************************
 ****************************************************************************/
void Shutdown() {
	DestroyPager();
	DestroyRootMenu();
	DestroyLoadDisplay();
	DestroyTray();
	ReleaseClients();
	ReleaseCursors();
	ReleaseFonts();
	ReleaseColors();
	XCloseDisplay(display);
}


